/*
** Control logic for Goldilocks DAC
** www.myprius.co.za
** 2012, AJ BREDEKAMP
*/

#include <p33fj128gp802.h>

#define FIFO_SIZE 2

//Declare Variables------------------------------------------------------------------------------------------------------------------------
#define _FAR __attribute__(( far))

//circular buffers
unsigned _FAR int FIFO_LEFT[FIFO_SIZE]; 	//buffer holding LEFT samples
unsigned _FAR int FIFO_RIGHT[FIFO_SIZE]; 	//buffer holding RIGHT samples

//FIFO write and read pointers
volatile int FIFO_W,FIFO_R;			 		//both FIFO_LEFT and FIFO_RIGHT will use the same pointers to make sure they remain synchronized
//-----------------------------------------------------------------------------------------------------------------------------------------

//OC1 ISR *NOT USED*
void __attribute__((__interrupt__, no_auto_psv)) _OC1Interrupt( void )
	{
		/* Interrupt Service Routine code goes here */
	IFS0bits.OC1IF = 0; // Clear OC1 interrupt flag
	}

//OC2 ISR *NOT USED*
void __attribute__((__interrupt__, no_auto_psv)) _OC2Interrupt( void )
	{
		/* Interrupt Service Routine code goes here */
	IFS0bits.OC1IF = 0; // Clear OC1 interrupt flag
	}

//Timer2 ISR *NOT USED*
void __attribute__((__interrupt__, no_auto_psv)) _T2Interrupt(void)
	{
		/* Interrupt Service Routine code goes here */
	IFS0bits.T2IF = 0; // Clear Timer3 Interrupt Flag
	}

//Timer3 ISR *NOT USED*
void __attribute__((__interrupt__, no_auto_psv)) _T3Interrupt(void)
	{
		/* Interrupt Service Routine code goes here */
        /* Initiate SPI write to SAA7030 (one LEFT, one RIGHT) here with corresponding read
        Timer 3 interrupt should occur somewhere after the STR1 pulse goes low. The left and right
        samples must be written immediately one after the other to their respective output buffers to minimize
        clock skew. Only one of the two SPI clocks will be used for both data streams. Data clock is 2.1168MHz, 16 times less
        that the instruction clock of 33.8688 MHz. Clock offset between SP1 and 2 is 29.525ns. Data bit period is 472.41ns.
        If SPI1 provides data and clock for LEFT samples and SPI2 provide only data for right samples, then the right data stream will
        be offset with the clock by 29.525ns. Given that the bit period is 472.41ns this variation should not be problematic.
        */

		//write to SPI1 Buffer - LEFT sample
        //write to SPI2 Buffer - RIGHT sample

        //increment or roll over FIFO_LEFT and RIGHT pointers. Can be performed by just XORing the FIFO pointer value with one. There
        //are only two positions, 0 and 1, in the FIFO/channel

	IFS0bits.T3IF = 0; // Clear Timer3 Interrupt Flag
	}

//DCI Interrupt handler recieves samples from RXBUF, processes them, and places them in the FIFO
void __attribute__((__interrupt__, no_auto_psv)) _DCIInterrupt(void)
    {													   
		/* Interrupt Service Routine code goes here */
        /* Read a sample from DCI buffer, XOR with 1000000000000000 and send to appropriate (L or R)FIFO output buffer */

	IFS3bits.DCIIF=0;  //Clear DCI Event Interrupt Flag
    } 

int main()
{ 

    // Initializations----------------------------------------------------------------------------------------

    //Set Oscillator Configuration for 33.8688 MIPS
	OSCCON = 0x0300;			// 01100000000 --> Primary oscillator (XT, HS, EC) with PLL
	CLKDIV = 0x0000;			// 
	PLLFBD = 0x001E;			// 11110 --> /32  PLLDIV
	
    // set the priority levels
	_IC1IP = 4;					// this is the default value anyway
    _DCIIP = 3;					// priority for DCI module
    _T2IP  = 4; 				// this is the default value anyway
	_T3IP  = 4; 				// this is the default value anyway
	_OC1IP = 4;					// this is the default value anyway
    _OC2IP = 4;					// this is the default value anyway
     
    // ---------------------------------------------------
    // ***************************************************
    // PERIPHERAL PINS------------------------------------
    // Configure Peripheral pin select registers
    // OC1 10010 RP8 tied to Output Compare 1, RP8R<4:0> --> CLOX
    _RP8R0 = 0;
	_RP8R1 = 1;
 	_RP8R2 = 0;
	_RP8R3 = 0;
	_RP8R4 = 1;
    // OC2 10011 RP7 tied to Output Compare 2, RP7R<4:0> --> STR1
    _RP7R0 = 1;
	_RP7R1 = 1;
 	_RP7R2 = 0;
	_RP7R3 = 0;
	_RP7R4 = 1;
    //-------------------------------------------------------------------------------------------------------------------- 

    // SPI1 Data Input 			SDI1 RPINR20 SDI1R<4:0> -- *NOT USED*
    // SPI1 Clock Input 		SCK1 RPINR20 SCK1R<4:0> -- *NOT USED*
    // SPI1 Slave Select Input 	SS1  RPINR21 SS1R<4:0>  -- *NOT USED*
 
    // SDO1 00111 RPn tied to SPI1 Data Output 			--> LEFT Channel DATA Out
    _RP2R0 = 1;
	_RP2R1 = 1;
 	_RP2R2 = 1;
	_RP2R3 = 0;
	_RP2R4 = 0;
    // SCK1 01000 RPn tied to SPI1 Clock Output         --> Data clock out
    _RP6R0 = 0;
	_RP6R1 = 0;
 	_RP6R2 = 0;
	_RP6R3 = 1;
	_RP6R4 = 0;   
    // SS1  01001 RPn tied to SPI1 Slave Select Output  -- *NOT USED*
    //--------------------------------------------------------------------------------------------------------------------

    // SPI2 Data  Input 		SDI2 RPINR22 SDI2R<4:0> -- *NOT USED*
    // SPI2 Clock Input 		SCK2 RPINR22 SCK2R<4:0> -- *NOT USED*
    // SPI2 Slave Select Input 	SS2  RPINR23 SS2R<4:0>  -- *NOT USED*
    
    // SDO2 01010 RPn tied to SPI2 Data Output 			--> RIGHT Channel DATA Out
    _RP3R0 = 0;
	_RP3R1 = 1;
 	_RP3R2 = 0;
	_RP3R3 = 1;
	_RP3R4 = 0;
    // SCK2 01011 RPn tied to SPI2 Clock Output         --> Data clock out *NOT USED* Both LEFT and RIGHT uses the same clock from SPI1 module RP6
    _RP5R0 = 1;
	_RP5R1 = 1;
 	_RP5R2 = 0;
	_RP5R3 = 1;
	_RP5R4 = 0;
    // SS2  01100 RPn tied to SPI2 Slave Select Output  -- *NOT USED*
    //--------------------------------------------------------------------------------------------------------------------

    // DCI Serial Data Input 	CSDI RPINR24 CSDIR<4:0>  - RP11 <-- DOUT  from DIR9001

    // DCI Serial Clock Input 	CSCK RPINR24 CSCKR<4:0>  - RP10 <-- BCKO  from DIR9001

    // DCI Frame Sync Input 	COFS RPINR25 COFSR<4:0>  - RP9  <-- LRCKO from DIR9001

    // ----------------------------------------------------
    // ****************************************************
    // OUTPUT COMPARE MODULES------------------------------
    // Setup Output Compare Module 1 for Toggle Mode -- OUTPUT = Fs x 384 = 16934400 Hz/4 = 4233600 Hz
    // Used to clock SAA7030 Digital Filter --> CLOX
    // Associated with Timer 2
	OC1CONbits.OCM = 0b000; 	// Disable Output Compare Module
	OC1CONbits.OCM = 0b010; 	// Define Initial State for OC1 Pin (High if OCM=0b010) 
	OC1CONbits.OCM = 0b000; 	// Disable Output Compare Module
	OC1CONbits.OCTSEL = 0; 		// Select Timer2 as output compare time base
	OC1R = 0; 					// Load the Compare Register Value 
	IPC0bits.OC1IP = 0x01; 		// Set Output Compare 1 Interrupt Priority Level
	IFS0bits.OC1IF = 0; 		// Clear Output Compare 1 Interrupt Flag 
	IEC0bits.OC1IE = 0; 		// Disable Output Compare 1 interrupt
	OC1CONbits.OCM = 0b011; 	// Select the Output Compare mode, Toggle Mode

	// Setup Output Compare Module 2 in Continuous Pulse mode
    // Used for SAA7030 STR1 strobe, frequency 44.1kHz. Rising edge follows completion of SAA7030 input data stream
    // Associated with Timer 3
	OC2CONbits.OCM = 0b000; 	// Disable Output Compare Module
	OC2CONbits.OCTSEL = 1; 		// Select Timer 3 as output compare time base
	OC2R = 0; 			    	// Load the Compare Register Value for rising edge
	OC2RS = 50; 				// Load the Compare Register Value for falling edge
	IPC1bits.OC2IP = 0x01; 		// Set Output Compare 2 Interrupt Priority Level
	IFS0bits.OC2IF = 0; 		// Clear Output Compare 2 Interrupt Flag  
	IEC0bits.OC2IE = 0; 		// Disable Output Compare 2 interrupt
	OC2CONbits.OCM = 0b101; 	// Select the Output Compare mode, Continuous Pulse mode
    // ----------------------------------------------------
    // ****************************************************
    // TIMERS----------------------------------------------
    // Initialize and enable Timer2
    // Clock Timer 2 from internal clock source Fosc
	T2CONbits.TON = 0; 			// Disable Timer
    T2CONbits.TCS = 0; 			// Select internal instruction cycle clock 
	T2CONbits.TGATE = 0; 		// Disable Gated Timer mode
	T2CONbits.TCKPS = 0b00; 	// Select 1:1 Prescaler
	TMR2 = 0x00; 				// Clear timer register
	PR2 = 7; 					// Load the period value, Timer is reset on period match 
                                // This value takes Fosc (33.8688MHz) and divide by 8 --> 4.2336MHz --> CLOX of SAA7030
	IPC1bits.T2IP = 0x01; 		// Set Timer2 Interrupt Priority Level
	IFS0bits.T2IF = 0; 			// Clear Timer2 Interrupt Flag

	// Initialize and enable Timer3
    // Clock Timer 3 from internal clock source Fosc
    T3CONbits.TON = 0; 			// Disable Timer
	T3CONbits.TCS = 0; 			// Select internal instruction cycle clock 
	T3CONbits.TGATE = 0; 		// Disable Gated Timer mode
	T3CONbits.TCKPS = 0b00; 	// Select 1:1 Prescaler
	TMR3 = 0x00; 				// Clear timer register
	PR3 = 1535; 				// Load the period value, Timer is reset on period match 
                                // This value takes Fosc (33.8688MHz) and divide by 768 --> 44.1kHz
	IPC2bits.T3IP = 0x01; 		// Set Timer3 Interrupt Priority Level
	IFS0bits.T3IF = 0; 			// Clear Timer3 Interrupt Flag
    // ----------------------------------------------------
    // ****************************************************
    // DCI-------------------------------------------------
    // Initialize Data Converter Interface
    // Reads sounds samples in from DIR9001
    /*
	In this section we will set up the DCI module for I2S operation to interface 
	with a DIR9001 SPDIF decoder receiving data at 44.1kHz STEREO The timing diagram is 
    provided in Fig 1 below:
                                  FIGURE 1
      _________________________________

      |		 	    	              |	                              |
COFS: |	  	     	                  |_______________________________|
       _   _   _   _   _   _   _   _   _   _   _   _   _   _   _   _   _    
CSCK:_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_|
          |<--------Right Channel Data--->|<----Left Channel Data-------->|
          |<---16bits---->|               |<---16bits---->|               |
          |<---RXBUF0---->|               |<---RXBUF0---->|               |
          |<--TimeSlot0-->|               |<--TimeSlot0-->|               |
          |<-----------1 Frame----------->|<-----------1 Frame----------->|
          |<--------------------1/Fs = 22.67 microseconds---------------->|
    */

	//DCI Control Register DCICON1 Initialization
	DCICON1bits.CSCKD 	= 1;	// Serial Bit Clock (CSCK pin) is input	DIR9001 controls data transfer
	DCICON1bits.CSCKE 	= 1;	// Data changes on falling edge sampled on rising edge of CSCK, from DIR9001 data sheet
	DCICON1bits.COFSD	= 1;	// Frame Sync Signal is input DIR9001 controls data transfer
	DCICON1bits.UNFM 	= 0;	// Transmit 0s on a transmit underflow
	DCICON1bits.CSDOM 	= 1;	// CSDO pin tristated during disabled transmit time slots, CSDO needs an EXTERNAL pull-down resistor
	DCICON1bits.DJST 	= 0;	// Data TX/RX is begun one serial clock cycle after frame sync pulse
	DCICON1bits.COFSM 	= 1;	// Frame Sync Signal set up for I2S mode 
	DCICON1bits.DLOOP 	= 0;	// No Loopback
	// DCI Control Register DCICON2 Initialization
	DCICON2bits.BLEN	= 0;	// One data word will be buffered between interrupts	
	DCICON2bits.COFSG 	= 0;	// Data frame has 1 word
	DCICON2bits.WS 		= 15;	// Data word size is 16 bits
	// DCI Control Register DCICON3 Initialization
	DCICON3 = 0;	    		// Set up CSCK Bit Clock Frequency (disabled)
	// Transmit Slot Control Register Initialization
	TSCONbits.TSE0 = 0;	    	// Transmit on Time Slot 0	DISABLED
	//Receiver Slot Control Register Initialization
	RSCONbits.RSE0 = 1;	    	// Receive on Time Slot 0	
	// Disable DCI Interrupt and Enable DCI module
	IFS3bits.DCIIF=0;  			//DCIIF: DCI Event Interrupt Flag Status bit
    IEC3bits.DCIIE=1; 			//DCIIE: DCI Event Interrupt Enable bit
	
    // ----------------------------------------------------
    // ****************************************************
    // SPI-------------------------------------------------
    // Initialize and enable Serial Peripheral Interface
    // Write sounds samples out to SAA7030
	SPI1CON1 = 0x143D;			// Primary Prescale 16:1,Secondary prescale 1:1,Master Mode ENABLE,16 bit mode, / 16 ; 10000111101
    SPI1CON2 = 0x0000;      	// LEFT
    SPI1STAT = 0x8000;     		// enable SPI port

    SPI2CON1 = 0x143D;			// Primary Prescale 16:1,Secondary prescale 1:1,Master Mode ENABLE,16 bit mode, / 16 ; 10000111101
    SPI2CON2 = 0x0000;      	// RIGHT
    SPI2STAT = 0x8000;     		// enable SPI port
    //-----------------------------------------------------

    //Activate Timers
    T2CONbits.TON = 1; 			// Start Timer2
	T3CONbits.TON = 1; 			// Start Timer3

    //Activate DCI Interrupt
    DCICON1bits.DCIEN = 1; 		// Enable DCI Module

    //OC Modules
    IEC0bits.T2IE = 0; 			// Disable Timer2 interrupt
    IEC0bits.T3IE = 0; 			// Disable Timer3 interrupt

    FIFO_W = 0;					// initialize pointer to ZERO
    FIFO_R = 0;					// initialize pointer to ZERO

	while(1)
	{   
   		//Nothing is done here. Everything is interrupt driven
	} 

	return 0;
	
} // main
//END --------------------------------------------------------------------------------------------------------
